/*
 *  This file is part of the nuvtools package
 *  Copyright (C) 2004, 2005 Gavin Hurlbut
 *
 *  nuvtools is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*HEADER---------------------------------------------------
* $Id: decode.c 105 2005-04-24 03:10:16Z gjhurlbu $
*
* Copyright 2004, 2005 Gavin Hurlbut
* All rights reserved
*
*----------------------------------------------------------
*/

#include "environment.h"
#include <stdlib.h>
#include "format.h"
#include "nuvread.h"
#include "avcodec.h"
#include "minilzo.h"
#include "RTjpeg.h"
#include "types.h"
#include "logging.h"
#include <lame/lame.h>

/* CVS generated ID string */
static char ident[] _UNUSED_ = 
    "$Id: decode.c 105 2005-04-24 03:10:16Z gjhurlbu $";

static lame_global_flags *gf;
static unsigned long int rtjpeg_table[128];
static RTjpeg_t *rtjpeg = NULL;
static char *rtjpeg_buf = NULL;


int findDecode( fourccDef_t *fourCCTable, int count, unsigned char frameType,
                unsigned int fourcc )
{
    int             i;

    for( i = 0; i < count; i++ ) {
        if( fourCCTable[i].frameType == frameType &&
            fourCCTable[i].fourCC.numeric == fourcc ) {
            return( i );
        }
    }
    return( -1 );
}


nuvFrame *decodeLzo( nuvFile *nv, nuvFrame *frameIn );

nuvFrame *decodeMpeg4( nuvFile *nv, nuvFrame *frameIn )
{
    frameIn->needsEncode = 0;
    return( frameIn );
}

nuvFrame *decodeMjpeg( nuvFile *nv, nuvFrame *frameIn )
{
    nuv_free_frame( frameIn );
    return( NULL );
}

nuvFrame *decodeRtjpeg( nuvFile *nv, nuvFrame *frameIn )
{
    nuvFrame       *frameOut;
    unsigned char  *planes[3];
    int             frameSize;

    frameOut = nuv_duplicate_frame(frameIn);
    if( !frameOut ) {
        nuv_free_frame( frameIn );
        return( NULL );
    }

    free(frameOut->data);

    frameOut->needsEncode = TRUE;
    frameOut->data = (char *)malloc(nv->frameSize);
    frameOut->length = nv->frameSize;

    frameSize = nv->width * nv->height;
    planes[0] = frameOut->data;
    planes[1] = planes[0] + frameSize;
    planes[2] = planes[1] + (frameSize / 4);
    RTjpeg_decompress(rtjpeg, frameIn->data, planes);
    
    nuv_free_frame( frameIn );
    return( frameOut );
}

nuvFrame *decodeRtjpegLzo( nuvFile *nv, nuvFrame *frameIn )
{
    nuvFrame       *frameOut;

    frameIn = decodeLzo( nv, frameIn );
    if( frameIn ) {
        frameOut = decodeRtjpeg( nv, frameIn );
    }
    return( frameOut );
}

nuvFrame *decodeRaw( nuvFile *nv, nuvFrame *frameIn )
{
    nuvFrame       *frameOut;

    frameOut = frameIn;
    frameOut->needsEncode = 1;
    return( frameOut );
}

nuvFrame *decodeRawLzo( nuvFile *nv, nuvFrame *frameIn )
{
    nuvFrame       *frameOut;

    frameOut = decodeLzo( nv, frameIn );
    if( frameOut ) {
        frameOut = decodeRaw( nv, frameIn );
    }
    return( frameOut );
}

nuvFrame *decodeLzo( nuvFile *nv, nuvFrame *frameIn )
{
    nuvFrame       *frameOut;
    int             ret;

    frameOut = nuv_duplicate_frame(frameIn);
    if( !frameOut ) {
        nuv_free_frame( frameIn );
        return( NULL );
    }

    free(frameOut->data);

    frameOut->needsEncode = 1;
    frameOut->data = (char *)malloc(nv->frameSize + 64);
    frameOut->length = nv->frameSize + 64;

    ret = lzo1x_decompress_safe(frameIn->data, frameIn->length, frameOut->data,
                                &(frameOut->length), NULL);
    nuv_free_frame( frameIn );
    if( ret != LZO_E_OK ) {
        LogPrint( LOG_CRIT, "LZO decode: %d\n", ret );
        nuv_free_frame( frameOut );
        return( NULL );
    }

    return( frameOut );
}

void decodeRtjpegInit( nuvFile *nv )
{
    int             val;
    
    rtjpeg = RTjpeg_init();
    RTjpeg_set_quality(rtjpeg, &nv->rtjpeg_quality);
    val = RTJ_YUV420;
    RTjpeg_set_format(rtjpeg, &val);
    RTjpeg_set_size(rtjpeg, &nv->width, &nv->height);
    val = 2;
    RTjpeg_set_intra(rtjpeg, &val, &nv->rtjpeg_luma, &nv->rtjpeg_chroma);

    rtjpeg_buf = malloc(nv->frameSize);
    memset(rtjpeg_buf, 0x00, nv->frameSize);
}

void writeFrameToDisk(nuvFrame *frame, int count)
{
    FILE           *fp;
    char            fileName[128];

    sprintf( fileName, "/tmp/frame%07d.yuv", count );

    fp = fopen( fileName, "wb" );
    fwrite( frame->data, frame->length, 1, fp );
    fclose( fp );
}

void decodeAudioInit( nuvFile *nv )
{
    gf = lame_init();
    lame_set_decode_only(gf, 1);
    lame_decode_init();
    lame_init_params(gf);
}

void decodeAudioClose( void )
{
    lame_close(gf);
}

nuvFrame *decodeMPx( nuvFile *nv, nuvFrame *frameIn )
{
    int         count;
    int         len;
    short int  *bufL;
    short int  *bufR;

    /* Leave space for 100ms of audio, although this SHOULD be overkill */
    len = nv->audio_sample_rate * sizeof(short int);

    bufL = (short int *)malloc(len);
    bufR = (short int *)malloc(len);

    if( !bufL || !bufR ) {
        if( bufL ) {
            free( bufL );
        }

        if( bufR ) {
            free( bufR );
        }
        nuv_free_frame( frameIn );
        return( NULL );
    }

    count = lame_decode( frameIn->data, frameIn->length, bufL, bufR );
    if( count <= 0 ) {
        nuv_free_frame( frameIn );
        free( bufL );
        free( bufR );
        return( NULL );
    }

    free( frameIn->data );
    if( frameIn->data2 ) {
        free( frameIn->data2 );
    }
    
    frameIn->data = (char *)bufL;
    frameIn->length = count * sizeof(short int);
    frameIn->data2 = (char *)bufR;
    frameIn->length2 = count * sizeof(short int);

    /* Mark as un-interlaced data */
    frameIn->needsEncode = 2;
    return( frameIn );
}

nuvFrame *decodePCM( nuvFile *nv, nuvFrame *frameIn )
{
    /* Mark as interlaced data */
    frameIn->needsEncode = 1;
    return( frameIn );
}


/*
 * vim:ts=4:sw=4:ai:et:si:sts=4
 */
