/*
 *  This file is part of the nuvtools package
 *  Copyright (C) 2004, 2005 Gavin Hurlbut
 *
 *  nuvtools is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*HEADER---------------------------------------------------
* $Id: main.c 105 2005-04-24 03:10:16Z gjhurlbu $
*
* Copyright 2004, 2005 Gavin Hurlbut
* All rights reserved
*
*/

/* INCLUDE FILES */
#include "environment.h"
#include "queue.h"
#include "logging.h"
#include <stdio.h>
#include <pthread.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <sys/time.h>
#include <errno.h>
#include <getopt.h>

#include "avformat.h"
#include "avcodec.h"
#include "types.h"
#include "release.h"

/* INTERNAL CONSTANT DEFINITIONS */

/* INTERNAL TYPE DEFINITIONS */

/* INTERNAL MACRO DEFINITIONS */
#define thread_create(a,b,c,d)  pthread_create((a),(b),(c),(d))

static pthread_t mainThreadId;
static pthread_t readerThreadId;
static pthread_t resyncThreadId;
static pthread_t videoReencodeThreadId;
static pthread_t audioReencodeThreadId;
static pthread_t writerThreadId;

static char *inputFileName;
static char *outputFileName;

QueueObject_t *ResyncQueue;
QueueObject_t *AudioInQueue;
QueueObject_t *VideoInQueue;
QueueObject_t *WritingQueue;

bool GlobalAbort = FALSE;

void *ReaderThread( void *arg );
void *ResyncThread( void *arg );
void *VideoReencodeThread( void *arg );
void *AudioReencodeThread( void *arg );
void *WriterThread( void *arg );

/* INTERNAL FUNCTION PROTOTYPES */
void LogBanner( void );
void MainParseArgs( int argc, char **argv );
void MainDisplayUsage( char *program, char *errorMsg );
void MainDelayReset( void );
void MainExit( int sigNum );

/* CVS generated ID string */
static char ident[] _UNUSED_= 
    "$Id: main.c 105 2005-04-24 03:10:16Z gjhurlbu $";

/*FUNCTION-------------------------------------------------
*
* Function Name :
* Returned Value :
* Parameters :
* Comments :
*
*----------------------------------------------------------
*/

int main ( int argc, char **argv )
{
    void *threadResult;
    int priority;
    int pid;

    GlobalAbort = FALSE;

    /* Parse the command line options */
    MainParseArgs( argc, argv );

    mainThreadId = pthread_self();

    /* Setup the exit handler */
    atexit( MainDelayReset );

    /* Setup the signal handler */
    signal( SIGINT, MainExit );

    /* Ignore SIGPIPE signals!  This allows TCP sockets to close while we're
     * writing to them.
     */
    signal( SIGPIPE, SIG_IGN );

    /* Print the startup log messages */
    LogBanner();

    /* Setup the Resync Queue */
    ResyncQueue = QueueCreate( 64 );

    /* Setup the Audio Input Queue */
    AudioInQueue = QueueCreate( 128 );

    /* Setup the Video Input Queue */
    VideoInQueue = QueueCreate( 128 );

    /* Setup the Writing Queue */
    WritingQueue = QueueCreate( 256 );

    
    /* Initialize the AVI stuff */
    avcodec_init();
    register_avcodec(&mp3lame_encoder);
    register_avcodec(&mpeg4_encoder);
    register_avcodec(&mjpeg_encoder);
    register_avcodec(&xvid_encoder);
    avienc_init();
    register_protocol(&file_protocol);

    /* Start the Reading thread */
    thread_create( &readerThreadId, NULL, ReaderThread, (void *)inputFileName );

    /* Start the Resync thread */
    thread_create( &resyncThreadId, NULL, ResyncThread, NULL );

    /* Start the Video Reencoding thread */
    thread_create( &videoReencodeThreadId, NULL, VideoReencodeThread, NULL );

    /* Start the Audio Reencoding thread */
    thread_create( &audioReencodeThreadId, NULL, AudioReencodeThread, NULL );

    /* Start the Writing thread */
    thread_create( &writerThreadId, NULL, WriterThread, 
                   (void *)outputFileName );

    /* Wait for Writing thread to finish */
    pthread_join( writerThreadId, &threadResult );

    /* Wait for Audio Reencoding thread to finish */
    pthread_join( audioReencodeThreadId, &threadResult );

    /* Wait for Video Reencoding thread to finish */
    pthread_join( videoReencodeThreadId, &threadResult );

    /* Wait for Resync thread to finish */
    pthread_join( resyncThreadId, &threadResult );

    /* Wait for Reading thread to finish */
    pthread_join( readerThreadId, &threadResult );

    return(0);
}


void LogBanner( void )
{
    LogPrint( LOG_CRIT, "\nnuv2avi  (c) 2004, 2005 Gavin Hurlbut\n" );
    LogPrint( LOG_CRIT, "%s\n", svn_version() );
}


void MainParseArgs( int argc, char **argv )
{
    extern char *optarg;
    extern int optind, opterr, optopt;
    int opt;
    int optIndex = 0;
    static struct option longOpts[] = {
        {"help", 0, 0, 'h'},
        {"version", 0, 0, 'V'},
        {"infile", 1, 0, 'i'},
        {"outfile", 1, 0, 'o'},
        {0, 0, 0, 0}
    };

    inputFileName   = NULL;
    outputFileName  = NULL;

    while( (opt = getopt_long( argc, argv, "hVi:o:", longOpts, &optIndex ))
           != -1 )
    {
        switch( opt )
        {
            case 'h':
                MainDisplayUsage( argv[0], NULL );
                exit( 0 );
                break;
            case 'i':
                if( inputFileName != NULL )
                {
                    free( inputFileName );
                }
                inputFileName = strdup(optarg);
                break;
            case 'o':
                if( outputFileName != NULL )
                {
                    free( outputFileName );
                }
                outputFileName = strdup(optarg);
                break;
            case 'V':
                LogBanner();
                exit( 0 );
                break;
            case '?':
            case ':':
            default:
                MainDisplayUsage( argv[0], "Unknown option" );
                exit( 1 );
                break;
        }
    }

    if( inputFileName == NULL || outputFileName == NULL )
    {
        MainDisplayUsage( argv[0], "You must specify both input and output "
                                   "files." );
        exit( 1 );
    }
}

void MainDisplayUsage( char *program, char *errorMsg )
{
    char *nullString = "<program name>";

    LogBanner();

    if( errorMsg != NULL )
    {
        fprintf( stderr, "\n%s\n\n", errorMsg );
    }

    if( program == NULL )
    {
        program = nullString;
    }

    fprintf( stderr, "\nUsage:\n\t%s -i inputFile -o outputFile\n\n",
             program );
    fprintf( stderr, "Options:\n"
                     "\t-i or --infile\tdefine the input .nuv file\n"
                     "\t-o or --outfile\tdefine the output .avi file\n"
                     "\t-V or --version\tshow the version number and quit\n"
                     "\t-h or --help\tshow this help text\n\n" );
}

void MainDelayReset( void )
{
    LogPrint( LOG_CRIT, "Shutting down\n" );

    /* Signal to all that we are aborting */
    GlobalAbort = TRUE;

    /* Delay so the threads can see the change */
    usleep( 300000 );

    /* Delay to allow all the other tasks to finish (esp. logging!) */
    usleep( 750000 );

#ifdef PROFILING
    _exit( 0 );
#endif
}

void MainExit( int sigNum )
{
    extern const char *const sys_siglist[];

    if( pthread_equal( pthread_self(), mainThreadId ) )
    {
        LogPrint( LOG_CRIT, "\nReceived signal: %s\n", sys_siglist[sigNum] );
        exit( 0 );
    }
}



/*
 * vim:ts=4:sw=4:ai:et:si:sts=4
 */

