/*
 *  This file is part of the nuvtools package
 *  Copyright (C) 2004, 2005 Gavin Hurlbut
 *
 *  nuvtools is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*HEADER---------------------------------------------------
* $Id: resync.c 105 2005-04-24 03:10:16Z gjhurlbu $
*
* Copyright 2004, 2005 Gavin Hurlbut
* All rights reserved
*
*/

#include "environment.h"
#include <stdio.h>
#include <stdlib.h>
#include "format.h"
#include "nuvread.h"
#include "queue.h"
#include "linked_list.h"
#include "avcodec.h"
#include "types.h"
#include "logging.h"

/* CVS generated ID string */
static char ident[] _UNUSED_ = 
    "$Id: resync.c 105 2005-04-24 03:10:16Z gjhurlbu $";

extern QueueObject_t *ResyncQueue;
extern QueueObject_t *AudioInQueue;
extern QueueObject_t *VideoInQueue;
extern bool GlobalAbort;

static LinkedList_t *videoList;
static LinkedList_t *audioList;

static nuvFile        *nv;
static nuvFrame       *frame;
static nuvFrame       *frame2;
static int             VideoOffset = 0;
static int             VideoPrev;
static int             VideoLastPrev;
static int             VideoFrameTime;
static int             AudioPrev;
static linkedFrame    *item;
static LinkedListItem_t *currVid, *currAud;
static int             timecode;

void *ResyncThread( void *arg )
{
    int             done;
    int             syncFrame = 0;
    int             synched = 0;
    FILE           *fLog;

    done = FALSE;

    fLog = fopen( "/tmp/timestamp.log", "wt" );

    videoList = LinkedListCreate();
    audioList = LinkedListCreate();

    if( !videoList || !audioList ) {
        fprintf( stderr, "Couldn't create linked list\n" );
        GlobalAbort = TRUE;
        return( NULL );
    }

    while( !done ) {
        frame = (nuvFrame *)QueueDequeueItem( ResyncQueue, -1 );
        if( !frame || GlobalAbort ) {
            return( NULL );
        }

        switch( frame->header.frametype ) {
        case 'I':
            nv = (nuvFile *)frame->data;

            VideoFrameTime = (int)(1000.0 * (float)nv->fps.denom /
                                   (float)nv->fps.num);
            VideoPrev = -1 * VideoFrameTime;
            AudioPrev = -1 * VideoFrameTime;

            frame2 = nuv_duplicate_frame( frame );

            frame->header.comptype = 'A';
            QueueEnqueueItem( AudioInQueue, (QueueItem_t)frame );

            frame2->header.comptype = 'V';
            QueueEnqueueItem( VideoInQueue, (QueueItem_t)frame2 );
            break;
        case 'E':
            done = TRUE;
            break;
        case 'A':
            if( synched ) {
                item = (linkedFrame *)malloc(sizeof(linkedFrame));
                if( item ) {
                    item->frame = frame;
                    LinkedListAdd( audioList, (LinkedListItem_t *)item,
                                   UNLOCKED, AT_TAIL );
                }
            }
            break;
        case 'V':
            if( synched ) {
                item = (linkedFrame *)malloc(sizeof(linkedFrame));
                if( item ) {
                    item->frame = frame;
                    LinkedListAdd( videoList, (LinkedListItem_t *)item,
                                   UNLOCKED, AT_TAIL );
                }
            }
            break;
        case 'S':
            if( frame->header.comptype == 'V' ) {
                syncFrame = 1;
                synched = 1;
            } else {
                nuv_free_frame( frame );
            }
            break;
        default:
            nuv_free_frame( frame );
            break;
        }


        if( syncFrame || done ) {
            VideoLastPrev = VideoPrev - VideoOffset;

            currAud = audioList->head;
            for(currVid = videoList->head; currVid; currVid = currVid->next) {
                timecode = ((linkedFrame *)currVid)->frame->header.timecode;
                if( (timecode - VideoPrev) > (VideoFrameTime + 4) ) {
                    /*
                     * We need to update the offset.  Let's do all the Audio
                     * frames up to the current timestamp with the old offset
                     * first
                     */
                    while( (item = (linkedFrame *)currAud) &&
                           item->frame->header.timecode < timecode) {
                        item->frame->header.timecode -= VideoOffset;
                        if( item->frame->header.timecode < AudioPrev + 4 ) {
                            item->frame->offset = -1;
                        } else {
                            fprintf( fLog, "A: %ld %ld %ld\n",
                                     item->frame->header.timecode + VideoOffset, 
                                     VideoOffset, 
                                     item->frame->header.timecode );
                            item->frame->offset = VideoOffset;
                            AudioPrev = item->frame->header.timecode;
                        }
                        currAud = currAud->next;
                    }
                    VideoOffset += (timecode - VideoPrev - VideoFrameTime);
                }
                VideoPrev = timecode;

                item = (linkedFrame *)currVid;
                fprintf( fLog, "V: %ld %ld ",
                         item->frame->header.timecode, VideoOffset );
                item->frame->header.timecode -= VideoOffset;
                item->frame->offset = VideoOffset;
                fprintf( fLog, "%ld\n",
                         item->frame->header.timecode );
            }

            /* Finish offsetting the audio */
            for(; currAud; currAud = currAud->next) {
                item = (linkedFrame *)currAud;
                item->frame->header.timecode -= VideoOffset;
                if( item->frame->header.timecode < AudioPrev + 4 ) {
                    item->frame->offset = -1;
                } else {
                    fprintf( fLog, "A: %ld %ld %ld\n",
                             item->frame->header.timecode + VideoOffset, 
                             VideoOffset, item->frame->header.timecode );
                    item->frame->offset = VideoOffset;
                    AudioPrev = item->frame->header.timecode;
                }
            }

            /* Now output audio to the audio thread */    
            for(currAud = audioList->head; currAud; currAud = audioList->head) {
                item = (linkedFrame *)currAud;
                if( item && item->frame->offset != -1 ) {
                    QueueEnqueueItem( AudioInQueue, (QueueItem_t)item->frame );
                }
                LinkedListRemove( audioList, currAud, UNLOCKED );
                free( item );
            }

            /* Output the video */
            for(currVid = videoList->head; currVid; currVid = videoList->head) {
                item = (linkedFrame *)currVid;
                if( item ) {
                    QueueEnqueueItem( VideoInQueue, item->frame );
                }
                LinkedListRemove( videoList, currVid, UNLOCKED );
                free( item );
            }

            /* The S or E frame */
            frame2 = nuv_duplicate_frame( frame );

            frame->header.comptype = 'A';
            QueueEnqueueItem( AudioInQueue, (QueueItem_t)frame );

            frame2->header.comptype = 'V';
            QueueEnqueueItem( VideoInQueue, (QueueItem_t)frame2 );
            syncFrame  = 0;
        }
    }

    LinkedListDestroy( audioList );
    LinkedListDestroy( videoList );

    fclose( fLog );

    return( NULL );
}


/*
 * vim:ts=4:sw=4:ai:et:si:sts=4
 */
