/*
 *  This file is part of the nuvtools package
 *  Copyright (C) 2004, 2005 Gavin Hurlbut
 *
 *  nuvtools is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*HEADER---------------------------------------------------
* $Id: writer.c 105 2005-04-24 03:10:16Z gjhurlbu $
*
* Copyright 2004, 2005 Gavin Hurlbut
* All rights reserved
*
*/

#ifndef _BSD_SOURCE
#define _BSD_SOURCE
#endif

#include "environment.h"
#include <stdlib.h>
#include <time.h>
#include <sys/time.h>
#include "format.h"
#include "avformat.h"
#include "nuvread.h"
#include "queue.h"
#include "types.h"
#include "logging.h"

/* CVS generated ID string */
static char ident[] _UNUSED_ = 
    "$Id: writer.c 105 2005-04-24 03:10:16Z gjhurlbu $";

extern QueueObject_t *WritingQueue;
extern bool GlobalAbort;

AVStream       *add_video_stream(AVFormatContext * oc, nuvFile *nv);
AVStream       *add_audio_stream(AVFormatContext * oc, nuvFile *nv);
AVFormatContext *av_alloc_format_context(void);


AVStream       *add_video_stream(AVFormatContext * oc, nuvFile *nv)
{
    AVCodecContext *c;
    AVStream       *st;

    st = av_new_stream(oc, 0);
    if (!st) {
        fprintf(stderr, "Could not alloc stream\n");
        exit(1);
    }

    c = &st->codec;
    c->codec_id = oc->oformat->video_codec;
    c->codec_type = CODEC_TYPE_VIDEO;
    c->codec_tag = nv->out_video_4cc;

    c->bit_rate        = nv->video_bitrate;
    c->width           = nv->width;
    c->height          = nv->height;
    c->frame_rate      = nv->fps.num;
    c->frame_rate_base = nv->fps.denom;
    c->gop_size        = nv->gop_length;

    return st;
}

AVStream       *add_audio_stream(AVFormatContext * oc, nuvFile *nv)
{
    AVCodecContext *c;
    AVStream       *st;

    st = av_new_stream(oc, 1);
    if (!st) {
        fprintf(stderr, "Could not alloc stream\n");
        exit(1);
    }

    c = &st->codec;
    c->codec_id = oc->oformat->audio_codec;
    c->codec_type = CODEC_TYPE_AUDIO;

    c->bit_rate    = nv->audio_bitrate_closest;
    c->sample_rate = nv->audio_sample_rate; 
    c->channels    = nv->audio_channels;

    return st;
}


void *WriterThread( void *arg )
{
    char           *filename;
    int             i;
    AVFormatContext *oc;
    AVOutputFormat *fmt;
    AVStream       *audio_st, *video_st;
    AVPacket        pkt;
    AVCodec        *codec;
    int             syncFrame = -1;
    nuvFile        *nv = NULL;
    nuvFrame       *frame;
    int             vidDone;
    int             audDone;
    long long int   frameCount;
    int             hrs, min, sec, usec;
    float           fps;
    int             Offset = 0;
    struct timeval  tvStart;
    struct timeval  tvNow;
    struct timeval  tv;
    struct timeval  tvLastDisp;
    long long int   diff;

    filename = (char *)arg;

    vidDone = TRUE;
    audDone = TRUE;
    while( vidDone || audDone ) {
        frame = (nuvFrame *)QueueDequeueItem( WritingQueue, -1 );
        if( frame->header.frametype == 'I' ) {
            if( !nv ) {
                nv = (nuvFile *)malloc( sizeof(nuvFile) );
                if( !nv ) {
                    LogPrint( LOG_CRIT, "Couldn't copy the file info in "
                                        "writer\n" );
                    return( NULL );
                }
                memcpy( nv, frame->data, frame->length );
            }

            if( frame->header.comptype == 'A' ) {
                audDone = FALSE;
            } else if( frame->header.comptype == 'V' ) {
                vidDone = FALSE;
            }

            free( frame );
        } else {
            nuv_free_frame(frame);
        }
    }

    fmt = guess_format("avi", NULL, NULL);
    oc = av_alloc_format_context();
    oc->oformat = fmt;
    snprintf(oc->filename, sizeof(oc->filename), "%s", filename);

    fmt->video_codec = codec_get_bmp_id(nv->out_video_4cc);
    fmt->audio_codec = CODEC_ID_MP3;

    video_st = add_video_stream(oc, nv);
    audio_st = add_audio_stream(oc, nv);

    codec = avcodec_find_encoder(video_st->codec.codec_id);
    avcodec_open(&video_st->codec, codec);
    codec = avcodec_find_encoder(audio_st->codec.codec_id);
    avcodec_open(&audio_st->codec, codec);

    av_set_parameters(oc, NULL);

    dump_format(oc, 0, filename, 1);

    url_fopen(&oc->pb, filename, URL_WRONLY);

    av_write_header(oc);

    frameCount = 0;

    gettimeofday( &tvStart, NULL ); 
    printf( "\n" );

    tvLastDisp.tv_sec = 0;
    tvLastDisp.tv_usec = 0;

    while( !audDone || !vidDone ) {
        frame = (nuvFrame *)QueueDequeueItem( WritingQueue, 1000 );
        gettimeofday( &tvNow, NULL );

        if( GlobalAbort ) {
            return( NULL );
        }

#if 0
        LogPrint( LOG_CRIT, "Frame Type %c\n", frame->header.frametype );
#endif

        if( !frame || (frameCount & 127) == 1LL ||
            frame->header.frametype == 'E' ) {
            timersub(&tvNow, &tvLastDisp, &tv);
            if( tv.tv_sec || frame->header.frametype == 'E' ) {
                tvLastDisp = tvNow;
                timersub(&tvNow, &tvStart, &tv);

                usec = tv.tv_usec;
                hrs  = tv.tv_sec / 3600;
                min  = (tv.tv_sec % 3600) / 60;
                sec  = tv.tv_sec % 60;
                
                if( !usec && !sec && !min && !hrs ) {
                    fps = 0.0;
                } else {
                    fps = (frameCount * 1.0) / 
                          ((tv.tv_sec * 1.0) + (tv.tv_usec * 0.000001));
                }
                
                printf( "\rFrame %9lld  Elapsed Time %02d:%02d:%02d.%06d  "
                        "Offset %5d.%03d  FPS %9.2f", 
                        frameCount, hrs, min, sec, usec, 
                        (Offset / 1000), abs(Offset % 1000), fps );
                fflush( stdout );
            }
        }

        if( !frame ) {
            continue;
        }

        switch( frame->header.frametype ) {
        case 'A':
            Offset = frame->offset;
            av_init_packet(&pkt);
            pkt.pts = ((int64_t)(frame->header.timecode)) * 1000;
            pkt.data = frame->data;
            pkt.size = frame->length;
            pkt.stream_index = audio_st->index;
            pkt.flags |= PKT_FLAG_KEY;
            av_interleaved_write_frame(oc, &pkt);
            break;
        case 'V':
            frameCount++;
            Offset = frame->offset;
            av_init_packet(&pkt);
            pkt.pts = ((int64_t)(frame->header.timecode)) * 1000;
            pkt.data = frame->data;
            pkt.size = frame->length;
            pkt.stream_index = video_st->index;
            if( syncFrame == 1 ) {
                pkt.flags |= PKT_FLAG_KEY;
                syncFrame = 0;
            }
            av_interleaved_write_frame(oc, &pkt);
            break;
        case 'S':
            syncFrame = 1;
            break;
        case 'E':
            if( frame->header.comptype == 'A' ) {
                audDone = TRUE;
            } else if( frame->header.comptype == 'V' ) {
                vidDone = TRUE;
            }
            break;
        }
        nuv_free_frame(frame);
    }

    printf( "\n\nTotal AVI Frame Count Written: %lld\n\n", frameCount );

    nuv_free_file(nv);

    av_write_trailer(oc);

    /* free the streams */
    for(i = 0; i < oc->nb_streams; i++) {
        av_freep(&oc->streams[i]);
    }

    url_fclose(&oc->pb);
    av_free(oc);

    return( NULL );
}



/*
 * vim:ts=4:sw=4:ai:et:si:sts=4
 */
